//
// $Id: App.cc 37 2005-06-18 06:05:24Z raphael $
//

#include <GL/freeglut.h>
#include <sstream>
#include <iostream>

#include "App.hh"

App::App(Config *cfg) {
  _config = cfg;
  _instance = this;
}

void App::run(int argc, char *argv[]) {
  _effect = new Effect(_config);
  initOpengl(argc, argv);
  _effect->init();
  _start = time(NULL);
  _frame = 0;
  glutMainLoop();
}

void App::end() {
  if(_config->isFullscreen()) {
    glutLeaveGameMode();
  } else {
    glutDestroyWindow(glutGetWindow());
  }
  glutLeaveMainLoop();
  long sec = time(NULL) - _start;
  std::cout << sec << " seconds, " << _frame << " frames ";
  std::cout << " -> " << _frame/sec << " fps" << std::endl;
}

void App::initOpengl(int argc, char *argv[]) {
  glutInit(&argc, argv);
  glutSetOption(
    GLUT_ACTION_ON_WINDOW_CLOSE,
    GLUT_ACTION_CONTINUE_EXECUTION
  );

  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

  if(_config->isFullscreen()) {
    std::ostringstream oss;
    oss << _config->getWidth() << "x" << _config->getHeight() << ":24@70";
    glutGameModeString(oss.str().c_str());
    glutEnterGameMode();
  } else {
    glutInitWindowSize(_config->getWidth(), _config->getHeight());
    glutCreateWindow("[title]");
  }

  glutDisplayFunc(wrapDisplayFunc);
  glutIdleFunc(wrapIdleFunc);
  glutReshapeFunc(wrapReshapeFunc);
  glutKeyboardFunc(wrapKeyboardFunc);
}

void App::displayFunc() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  _effect->displayFunc();
  glFlush();
  glutSwapBuffers();
  _frame++;
}

void App::idleFunc() {
  // has(timer) -> update
  glutPostRedisplay();
}

void App::reshapeFunc(int w, int h) {
  _config->resize(w, h);
  _effect->reshapeFunc(w, h);
}

void App::keyboardFunc(unsigned char key, int x, int y) {
  switch(key) {
  case 27:
  case 'q':
    end();
    break;
  case 'a':
    _effect->angleInc();
    break;
  case 's':
    _effect->angleDec();
    break;
  case 'c':
    _effect->particleRateDec();
    break;
  case 'd':
    _effect->particleRateInc();
    break;
  case 'r':
    _effect->resetWater();
    break;
  case 'w':
    _effect->wfWater();
    break;
  case 'v':
    _effect->particleSizeDec();
    break;
  case 'f':
    _effect->particleSizeInc();
    break;
  default:
    break;
  }
}

App *App::_instance = 0;

App *App::getInstance() {
  return _instance;
}

//
// opengl wrapper
//

void wrapDisplayFunc() {
  App::getInstance()->displayFunc();
}

void wrapIdleFunc() {
  App::getInstance()->idleFunc();
}

void wrapReshapeFunc(int w, int h) {
  App::getInstance()->reshapeFunc(w,h);

}

void wrapKeyboardFunc(unsigned char key, int x, int y) {
  App::getInstance()->keyboardFunc(key, x, y);
}



