//
// $Id: Effect.cc 43 2005-06-23 06:13:12Z raphael $
//

#include <sys/timeb.h>
#include <iostream>
#include <fstream>

#include "Effect.hh"

Effect::Effect(Config *cfg) {
  _config = cfg;
  _water = new Water();
  _system = new ParticleSystem(_water);
  _angle = 30;
  _tm_sec = 0;
  _tm_msec = 0;
  _model = 1;

  _dt = 1.0/70;
}

void Effect::init() {
  int size;
  GLubyte *buf;
  std::ifstream fp;

  fp.open("data/mountain.bmp", std::ios::in | std::ios::binary);
  fp.seekg(18, std::ios::beg);
  fp.read((char *)&size, 4);

  buf = new GLubyte[size*size*3];

  fp.seekg(54, std::ios::beg);
  fp.read((char *)buf, size*size*3);

  glGenTextures(1, &_texture);
  glBindTexture(GL_TEXTURE_2D, _texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, size, size, 0, GL_BGR, GL_UNSIGNED_BYTE, buf);

  fp.close();
  delete[] buf;

  _system->init();
  _water->init(_texture, _config->waterSize());

  initScene();

  GLfloat ambient[] = {0.1, 0.1, 0.1, 1.0};
  GLfloat diffuse[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat position[] = {1.0, 1.0, 1.0, 0.0};

  _mat_shininess[0] = 16.0;
  _mat_specular[0] = 0.9;
  _mat_specular[1] = 0.9;
  _mat_specular[2] = 0.9;
  _mat_specular[3] = 1.0;
  _mat_diffuse[0] = 0.5;
  _mat_diffuse[1] = 0.5;
  _mat_diffuse[2] = 0.5;
  _mat_diffuse[3] = 1.0;

  glEnable(GL_DEPTH_TEST);

  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, position);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  glMaterialfv(GL_FRONT, GL_SHININESS, _mat_shininess);
  glMaterialfv(GL_FRONT, GL_SPECULAR, _mat_specular);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, _mat_diffuse);

  glShadeModel(GL_SMOOTH);
  
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_NORMALIZE);
  
  glEnable(GL_CULL_FACE);
  glPolygonMode(GL_FRONT, GL_FILL);
//   glPolygonMode(GL_BACK, GL_LINE);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

  reshapeFunc(_config->getWidth(), _config->getHeight());

  struct timeb tp;
  ftime(&tp);
  _tm_sec = tp.time;
  _tm_msec = tp.millitm;
}

void Effect::initScene() {
  int r = 32;
  int h = 8;
  int b = 2;

  glNewList(_model, GL_COMPILE);
  glColor4f(.4, .4, .4, 1);
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 0, 1);
    glVertex3f(-r, h, -r);
    glVertex3f(-r, -h, -r);
    glVertex3f( r, h, -r);
    glVertex3f( r, -h, -r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 1, 0);
    glVertex3f(-r-b, h, -r-b);
    glVertex3f(-r-b, h, -r);
    glVertex3f( r+b, h, -r-b);
    glVertex3f( r+b, h, -r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 0, -1);
    glVertex3f(-r-b, -h, -r-b);
    glVertex3f(-r-b, h, -r-b);
    glVertex3f( r+b, -h, -r-b);
    glVertex3f( r+b, h, -r-b);
  } glEnd();

  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 0, -1);
    glVertex3f( r, h, r);
    glVertex3f( r, -h, r);
    glVertex3f(-r, h, r);
    glVertex3f(-r, -h, r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 1, 0);
    glVertex3f( r+b, h, r+b);
    glVertex3f( r+b, h, r);
    glVertex3f(-r-b, h, r+b);
    glVertex3f(-r-b, h, r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 0, 1);
    glVertex3f( r+b, -h, r+b);
    glVertex3f( r+b, h, r+b);
    glVertex3f(-r-b, -h, r+b);
    glVertex3f(-r-b, h, r+b);
  } glEnd();

  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(1, 0, 0);
    glVertex3f(-r, h,  r);
    glVertex3f(-r, -h,  r);
    glVertex3f(-r, h, -r);
    glVertex3f(-r, -h, -r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 1, 0);
    glVertex3f(-r-b, h,  r);
    glVertex3f(-r,   h,  r);
    glVertex3f(-r-b, h, -r);
    glVertex3f(-r,   h, -r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(-1, 0, 0);
    glVertex3f(-r-b, -h,  r+b);
    glVertex3f(-r-b, h,  r+b);
    glVertex3f(-r-b, -h, -r-b);
    glVertex3f(-r-b, h, -r-b);
  } glEnd();

  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(-1, 0, 0);
    glVertex3f(r, h, -r);
    glVertex3f(r, -h, -r);
    glVertex3f(r, h,  r);
    glVertex3f(r, -h,  r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 1, 0);
    glVertex3f(r+b, h, -r);
    glVertex3f(r,   h, -r);
    glVertex3f(r+b, h,  r);
    glVertex3f(r,   h,  r);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(1, 0, 0);
    glVertex3f(r+b, -h, -r-b);
    glVertex3f(r+b, h, -r-b);
    glVertex3f(r+b, -h,  r+b);
    glVertex3f(r+b, h,  r+b);
  } glEnd();
  glBegin(GL_TRIANGLE_STRIP); {
    glNormal3f(0, 1, 0);
    glVertex3f(-r, -h, -r);
    glVertex3f(-r, -h,  r);
    glVertex3f( r, -h, -r);
    glVertex3f( r, -h,  r);
  } glEnd();

  glEndList();
}

void Effect::background() {
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glBindTexture(GL_TEXTURE_2D, _texture);

  glColor4f(.7, .7, .7, 1);

  glBegin(GL_TRIANGLE_FAN); {
    glTexCoord2f(0.0, 0.0);
    glVertex2f(-1.0, -1.0);
    glTexCoord2f(1.0, 0.0);
    glVertex2f(1.0, -1.0);
    glTexCoord2f(1.0, 1.0);
    glVertex2f(1.0, 1.0);
    glTexCoord2f(0.0, 1.0);
    glVertex2f(-1.0, 1.0);
  } glEnd();

  glDisable(GL_TEXTURE_2D);
}

void Effect::displayFunc() {
  struct timeb tp;
  ftime(&tp);
  float dt = (tp.time - _tm_sec) + (tp.millitm - _tm_msec)/1000.0;
  _tm_sec = tp.time;
  _tm_msec = tp.millitm;

  dt = (dt > _dt) ? _dt : dt;

  _system->update(dt);
  _water->update(dt);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.0, 1.0, -1.0, 1.0, 0.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glPushMatrix();
  background();
  glClear(GL_DEPTH_BUFFER_BIT);
  glPopMatrix();

  glViewport(0, 0, _w, _h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-_frustrum, _frustrum, -1.0, 1.0, 1.0, 100.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0, -20, -44);
  glRotatef(_angle, 0, 1, 0);

  glPushMatrix();

  glPushMatrix();
  glCallList(_model);
  glPopMatrix();

  glPushMatrix();
  _water->draw();
  glPopMatrix();

  glPushMatrix();
  _system->draw();
  glPopMatrix();

  glPopMatrix();
}

void Effect::reshapeFunc(int w, int h) {
  _frustrum = w / (float)h;
  _w = w, _h = h;

  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-_frustrum, _frustrum, -1.0, 1.0, 1.0, 100.0);
}

void Effect::angleInc() {
  _angle += 1;
}

void Effect::angleDec() {
  _angle -= 1;
}

void Effect::particleRateInc() {
  _system->rateInc();
}

void Effect::particleRateDec() {
  _system->rateDec();
}

void Effect::resetWater() {
  _water->reset();
}

void Effect::wfWater() {
  _water->toggleWF();
}

void Effect::particleSizeInc() {
  _system->sizeInc();
}

void Effect::particleSizeDec() {
  _system->sizeDec();
}

