//
// $Id: Water.cc 43 2005-06-23 06:13:12Z raphael $
//

#include <cstdlib>
#include <cmath>

#include "Water.hh"

Water::Water() {
}

void Water::init(GLuint bgTex, int w) {
  int i;

  _width = w;

  _texture = bgTex;
  _wireframe = 0;

  _c = .1;
  _k = .0002;
  _h = 1/float(_width);

  _map = new WaterElem*[_width];
  for(i=0; i<_width; i++) {
    _map[i] = new WaterElem[_width];
  }

  reset();
}

void Water::reset() {
  int i, j;
  for(j=0;j<_width;j++) {
    float fy = (2.0*j)/(_width-1)-1;
    for(i=0;i<_width;i++) {
      float fx = ((2.0*i)/(_width-1)-1);
      _map[i][j].u[0] = 32*fx;
      _map[i][j].u[2] = 32*fy;
      _map[i][j].u[1] = 0;
      _map[i][j].n[0] = 0;
      _map[i][j].n[1] = 1;
      _map[i][j].n[2] = 0;
      _map[i][j].v = 0;
    }
  }
}

void Water::update(float dt) {
  int i, j;
  float u;
  float h2 = _h*_h;
  // wave
  for(j=1;j<_width-1;j++) {
    for(i=1;i<_width-1;i++) {
      u =  _map[i+1][j  ].u[1] + _map[i-1][j  ].u[1];
      u += _map[i  ][j+1].u[1] + _map[i  ][j-1].u[1];
      u -= 4 * _map[i][j].u[1];
      _map[i][j].v =  _map[i][j].v + dt*_c*_c*u/h2;
    }
  }
  for(j=1;j<_width-1;j++) {
    for(i=1;i<_width-1;i++) {
      _map[i][j].u[1] = _map[i][j].u[1] + dt*_map[i][j].v;
    }
  }
  // bound
  for(i=0;i<_width;i++) {
    int w = _width-1;
    _map[i][0].u[1] = _map[i][1  ].u[1];
    _map[i][w].u[1] = _map[i][w-1].u[1];
    _map[0][i].u[1] = _map[1  ][i].u[1];
    _map[w][i].u[1] = _map[w-1][i].u[1];
  }

  // diffusion
  for(j=1;j<_width-1;j++) {
    for(i=1;i<_width-1;i++) {
      u =  _map[i+1][j].u[1] + _map[i-1][j].u[1];
      u += _map[i][j+1].u[1] + _map[i][j-1].u[1];
      u -= 4*_map[i][j].u[1];
      _map[i][j].u[1] += dt * _k * u / h2; 
    }
  }  

  for(j=1;j<_width-1;j++) {
    for(i=1;i<_width-1;i++) {
      float l[3];
      float m[3];
      l[0] = _map[i][j-1].u[0] - _map[i][j+1].u[0];
      l[1] = _map[i][j-1].u[1] - _map[i][j+1].u[1];
      l[2] = _map[i][j-1].u[2] - _map[i][j+1].u[2];
      m[0] = _map[i-1][j].u[0] - _map[i+1][j].u[0];
      m[1] = _map[i-1][j].u[1] - _map[i+1][j].u[1];
      m[2] = _map[i-1][j].u[2] - _map[i+1][j].u[2];
      _map[i][j].n[0] = l[1]*m[2]-l[2]*m[1];
      _map[i][j].n[1] = l[2]*m[0]-l[0]*m[2];
      _map[i][j].n[2] = l[0]*m[1]-l[1]*m[0];
    }
  }
}

void Water::draw() {
  glColor4f(1, 1, 1, .5);
  glPushMatrix(); {

    if(_wireframe > 0) {
      glPolygonMode(GL_FRONT, GL_LINE);
    }
    if(_wireframe < 2) {
      glEnable(GL_TEXTURE_2D);
    }

    glBindTexture(GL_TEXTURE_2D, _texture);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    
    glEnable(GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    for(int j=1;j<_width;j++) {
      glNormal3f(0, 1, 0);
      glBegin(GL_TRIANGLE_STRIP); {
	for(int i=0;i<_width;i++) {
	  glTexCoord2f(i/float(_width), (j-1)/float(_width));
          glNormal3fv(_map[i][j-1].n);
          glVertex3fv(_map[i][j-1].u);
	  glTexCoord2f(i/float(_width), j/float(_width));
	  glNormal3fv(_map[i][j].n);
	  glVertex3fv(_map[i][j].u);
	}
      } glEnd();
    }

    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_2D);

    if(_wireframe) {
      glPolygonMode(GL_FRONT, GL_FILL);
    }
  } glPopMatrix();
}

void Water::drop(GLfloat pos[3]) {
  int i, j;
  i = (int)(_width*(pos[0]+32)/64.0);
  j = (int)(_width*(pos[1]+32)/64.0);
  if((i>=0) && (i<_width) && (j>=0) && (j<_width)) {
    _map[i  ][j  ].u[1] -= 2;
    _map[i+1][j  ].u[1] -= 2;
    _map[i  ][j+1].u[1] -= 2;
    _map[i+1][j+1].u[1] -= 2;

    _map[i-1][j  ].u[1] += 1;
    _map[i-1][j+1].u[1] += 1;
    _map[i  ][j+2].u[1] += 1;
    _map[i+1][j+2].u[1] += 1;
    _map[i+2][j+1].u[1] += 1;
    _map[i+2][j  ].u[1] += 1;
    _map[i  ][j-1].u[1] += 1;
    _map[i+1][j-1].u[1] += 1;
  }
}

void Water::toggleWF() {
  _wireframe = (_wireframe + 1) % 3;
}
